#include <SPAD/DEV_KRNL.H>
#include <SPAD/BIO.H>

#include "SD.H"

void SCSI_QUEUE_PASS_THROUGH(SD *sd, IOCTLRQ *rq)
{
	VDESC vdesc;
	VPHYS vphys;

	if (__unlikely(rq->v.len < sizeof(SCSI_PASS_THROUGH_HEAD))) {
		einval_ret:
		rq->status = -EINVAL;
		call_ret:
		CALL_AST(rq);
		return;
	}

	if (__unlikely((unsigned long)rq->v.ptr & (BIO_SECTOR_SIZE - 1)))
		goto einval_ret;

	if (__unlikely((rq->v.len - sizeof(SCSI_PASS_THROUGH_HEAD) + BIO_SECTOR_SIZE - 1) >> BIO_SECTOR_SIZE_BITS > sd->ap.max_sectors)) {
		rq->status = -ERANGE;
		goto call_ret;
	}
	if (__unlikely(sd->flags & FLAGS_MEDIA_PROBING)) {
		WQ_WAIT_F(&sd->media_wq_noerror, rq);
		return;
	}
	if (__unlikely(sd->pass_through != NULL)) {
		WQ_WAIT_F(&sd->pass_through_wait, rq);
		return;
	}

	vdesc.ptr = rq->v.ptr;
	vdesc.len = sizeof(SCSI_PASS_THROUGH_HEAD);
	vdesc.vspace = rq->v.vspace;
	RAISE_SPL(SPL_VSPACE);
	rq->v.vspace->op->vspace_physlock(&vdesc, PF_RW, &vphys, &sd->pass_through_physunlock);
	LOWER_SPL(SPL_ATA_SCSI);
	if (!vphys.len) {
		DO_PAGEIN_NORET(rq, &rq->v, PF_RW);
		return;
	}
	sd->pass_through_phys = vphys.ptr;
	if (__unlikely(vphys.len != sizeof(SCSI_PASS_THROUGH_HEAD))) {
		RAISE_SPL(SPL_VSPACE);
		sd->pass_through_physunlock(sd->pass_through_phys);
		LOWER_SPL(SPL_ATA_SCSI);
		goto einval_ret;
	}

	sd->pass_through = rq;
	SD_DEQUEUE(&sd->ap);
}

int SCSI_PROCESS_PASS_THROUGH(SD *sd)
{
	SCSI_PASS_THROUGH_HEAD *pth;
	int r;
	if (__unlikely(sd->pass_through_posted))
		return 0;
	if (__unlikely(sd->free_tags != sd->max_tags))
		return 0;
	
	pth = KERNEL$MAP_PHYSICAL_BANK(sd->pass_through_phys);

	sd->pass_through_cmd.direction = pth->direction;
	if (!sd->pass_through_cmd.direction) {
		if (__unlikely(sd->pass_through->v.len != sizeof(SCSI_PASS_THROUGH_HEAD)))
			goto einval_unmap_ret;
	} else if (sd->pass_through_cmd.direction == PF_READ ||
		   __likely(sd->pass_through_cmd.direction == PF_WRITE)) {
		if (__unlikely(sd->pass_through->v.len == sizeof(SCSI_PASS_THROUGH_HEAD)))
			goto einval_unmap_ret;
	} else goto einval_unmap_ret;

	sd->pass_through_cmd.cmdlen = pth->command_len;
	if (__unlikely(sd->pass_through_cmd.cmdlen > SCSI_PASS_THROUGH_CMD_LEN)) {
		einval_unmap_ret:
		r = -EINVAL;
		KERNEL$UNMAP_PHYSICAL_BANK(pth);
		goto post_r_ret;
	}
	memcpy(sd->pass_through_cmd.cmd, pth->command, sd->pass_through_cmd.cmdlen);
	if (sd->scsi_version < 3) {
		sd->pass_through_cmd.cmd[1] &= ~0xe0;
		sd->pass_through_cmd.cmd[1] |= sd->lun_byte;
	}
	if (!sd->pass_through_cmd.direction) {
		sd->pass_through_cmd.desc = NULL;
	} else {
		sd->pass_through_cmd.desc = &sd->pass_through_desc;
		sd->pass_through_desc.v.ptr = sd->pass_through->v.ptr + sizeof(SCSI_PASS_THROUGH_HEAD);
		sd->pass_through_desc.v.len = sd->pass_through->v.len - sizeof(SCSI_PASS_THROUGH_HEAD);
		sd->pass_through_desc.v.vspace = sd->pass_through->v.vspace;
		sd->pass_through_desc.next = NULL;
	}
	sd->pass_through_cmd.internal = 0;
	KERNEL$UNMAP_PHYSICAL_BANK(pth);

	memset(&sd->pass_through_sense, 0, SCSI_PASS_THROUGH_SENSE_LEN);
	sd->pass_through_cmd.scsi_status = 0;

	r = sd->ap.post((SCSIRQ *)&sd->pass_through_cmd);
	if (__unlikely(r)) {
		if (r == -EAGAIN) {
			return 0;
		}
		if (r == -EVSPACEFAULT) {
			RAISE_SPL(SPL_VSPACE);
			sd->pass_through_physunlock(sd->pass_through_phys);
			LOWER_SPL(SPL_ATA_SCSI);
			DO_PAGEIN_NORET(sd->pass_through, &sd->pass_through->v, PF_RW);
		} else {
			post_r_ret:
			RAISE_SPL(SPL_VSPACE);
			sd->pass_through_physunlock(sd->pass_through_phys);
			LOWER_SPL(SPL_ATA_SCSI);
			sd->pass_through->status = r;
			CALL_AST(sd->pass_through);
		}
		sd->pass_through = NULL;
		WQ_WAKE_ALL(&sd->pass_through_wait);
		return 1;
	}
	sd->pass_through_posted = 1;
	KERNEL$SET_TIMER(sd->timeout, &sd->pass_through_timer);
	return 1;
}

void SCSI_PASS_THROUGH_DONE(SCSIRQ *cmd)
{
	SCSI_PASS_THROUGH_HEAD *pth;
	SD *sd = GET_STRUCT(cmd, SD, pass_through_cmd);

	sd->pass_through_posted = 0;
	KERNEL$DEL_TIMER(&sd->pass_through_timer);

	pth = KERNEL$MAP_PHYSICAL_BANK(sd->pass_through_phys);
	if (__likely(sd->pass_through_cmd.status >= 0)) {
		pth->scsi_status = sd->pass_through_cmd.scsi_status;
		if (__unlikely(pth->scsi_status == SCSI_CHECK_CONDITION)) {
			memcpy(pth->sense, sd->pass_through_sense, SCSI_PASS_THROUGH_SENSE_LEN);
		}
	}
	KERNEL$UNMAP_PHYSICAL_BANK(pth);

	RAISE_SPL(SPL_VSPACE);
	sd->pass_through_physunlock(sd->pass_through_phys);
	LOWER_SPL(SPL_ATA_SCSI);
	sd->pass_through->status = sd->pass_through_cmd.status;
	CALL_AST(sd->pass_through);

	sd->pass_through = NULL;
	WQ_WAKE_ALL(&sd->pass_through_wait);
	SD_DEQUEUE(&sd->ap);
}

void SCSI_PASS_THROUGH_TIMER_FN(TIMER *t)
{
	SD *sd = GET_STRUCT(t, SD, pass_through_timer);
	LOWER_SPL(SPL_ATA_SCSI);
	SET_TIMER_NEVER(&sd->pass_through_timer);
	sd->ap.cancel((SCSIRQ *)&sd->pass_through_cmd);
}
