#include <KERNEL/DEV.H>
#include <KERNEL/VMDEF.H>
#include <KERNEL/KSPAWN.H>
#include <SPAD/BIO_KRNL.H>
#include <SPAD/PKTSTUB.H>
#include <SPAD/IOCTL.H>
#include <SPAD/HASH.H>
#include <STDLIB.H>
#include <LIMITS.H>
#include <TIME.H>
#include <UNISTD.H>
#include <SPAD/LINK.H>
#include <SPAD/DL.H>
#include <DLFCN.H>
#include <KERNEL/LINK.H>
#include <SPAD/SYNC.H>

MTX_DECL(DEV_MUTEX, "KERNEL$DEV_MUTEX");

static void PENDING_DELETE(FFILE *f);
static void FREE_BLOB_REF(FBLOB_REF *br);

int UNREF_FILE(FFILE *f)
{
#if __DEBUG >= 1
	if (__unlikely(!f->l_refcount))
		KERNEL$SUICIDE("UNREF_FILE: FILE %p HAS ZERO REFCOUNT", f);
#endif
	if (__likely(!--f->l_refcount)) {
		FREE_PIPE(f);
		while (__unlikely(!XLIST_EMPTY(&f->blobs))) {
			FBLOB_REF *br = LIST_STRUCT(f->blobs.next, FBLOB_REF, list);
			FREE_BLOB_REF(br);
		}
		if (__unlikely(f->flags & _O_PENDING_DELETE)) {
			PENDING_DELETE(f);
		} else {
			free(f);
		}
		return 1;
	}
	return 0;
}

void FREE_PIPE_INTERNAL(FFILE *f)
{
	FPIPE *p = (FPIPE *)(unsigned long)f->pos;
	if (__likely(p != NULL) && !--p->l_refcount) {
		KERNEL$FREE_KERNEL_PAGE(p, VM_TYPE_WIRED_MAPPED);
	}
}

DECL_XLIST(PENDING_DELETE_LIST);
extern AST_STUB PENDING_DELETE_FN;

static void PENDING_DELETE(FFILE *f)
{
	static OPENRQ PENDING_DELETE_RQ;
	if (XLIST_EMPTY(&PENDING_DELETE_LIST)) {
		PENDING_DELETE_RQ.fn = PENDING_DELETE_FN;
		PENDING_DELETE_RQ.status = 1;
		CALL_AST(&PENDING_DELETE_RQ);
	}
	ADD_TO_XLIST(&PENDING_DELETE_LIST, (LIST_ENTRY *)(void *)f);
}

DECL_AST(PENDING_DELETE_FN, SPL_DEV, OPENRQ)
{
	FFILE *f;
	if (RQ->status != 1) {
		f = (FFILE *)PENDING_DELETE_LIST.next;
		DEL_FROM_LIST((LIST_ENTRY *)f);
		free(f);
		if (__likely(XLIST_EMPTY(&PENDING_DELETE_LIST))) RETURN;
	}
	f = (FFILE *)PENDING_DELETE_LIST.next;
	RQ->flags = _O_DELETE | _O_CLOSE | _O_NOPOSIX;
	RQ->path = f->path;
	RQ->cwd = NULL;
	RETURN_IORQ(RQ, KERNEL$OPEN);
}

static void FREE_BLOB_REF(FBLOB_REF *br)
{
	KERNEL$UNREF_BLOB(br->blob);
	DEL_FROM_LIST(&br->list);
	free(br);
}

void KERNEL$UNREF_BLOB(FBLOB *b)
{
	int spl = KERNEL$SPL;
	if (SPLX_BELOW(spl, SPL_X(SPL_DEV))) RAISE_SPL(SPL_DEV);
#if __DEBUG >= 1
	if (__unlikely(!b->l_refcount))
		KERNEL$SUICIDE("KERNEL$UNREF_BLOB: FILE'S BLOB %08X (AT %p) HAS ZERO REFCOUNT", b->type, b);
#endif
	if (__likely(!--b->l_refcount)) {
		free(b);
	}
	LOWER_SPLX(spl);
}

void SET_FILE_BLOB(FFILE *f, FBLOB_REF *br, FBLOB *bl, void **free1, void **free2)
{
	FBLOB_REF *bs;
	*free1 = *free2 = NULL;
	br->blob = bl;
	bl->l_refcount++;
	XLIST_FOR_EACH(bs, &f->blobs, FBLOB_REF, list) if (bs->blob->type == bl->type) {
		FBLOB *orig_blob = bs->blob;
		if (__unlikely(!bl->size)) {
			DEL_FROM_LIST_ORDERED(&bs->list);
			*free2 = bs;
			goto unref_free_br_ret;
		}
		bs->blob = br->blob;
		unref_free_br_ret:
#if __DEBUG >= 1
		if (__unlikely(!orig_blob->l_refcount))
			KERNEL$SUICIDE("SET_FILE_BLOB: FILE'S BLOB %08X (AT %p) HAS ZERO REFCOUNT", orig_blob->type, orig_blob);
#endif
		*free1 = !--orig_blob->l_refcount ? orig_blob : NULL;
		free_br_ret:
		free(br);
		return;
	}
	if (__unlikely(!bl->size)) goto free_br_ret;
	ADD_TO_XLIST_ORDERED(&f->blobs, &br->list);
	return;
}

DECL_IOCALL(KERNEL$NO_OPERATION, SPL_TOP, IORQ)
{
	/* We could do SWITCH_PROC_ACCOUNT here, but it's not needed because
	   this is called directly from process without blocking.
	   We can't call it because we are on SPL_TOP and we don't know which
	   request type we have (SIORQ, AIORQ, IOCTLRQ, BIORQ, PACKET) */
	RQ->status = -ENOOP;
	RETURN_AST(RQ);
}

static IORQ geth;
static VDESC geth_vdesc;
static WQ_DECL(geth_wait, "KERNEL$GETH_WAIT");
static IORQ * volatile geth_caller = NULL;
static int geth_wr;

extern IO_STUB NOOP_CONT_PAGEIN;

static IORQ *NOOP_GET_PAGEIN_RQ(VDESC *desc, IORQ *rq, int wr)
{
	if (__unlikely(geth_caller != NULL)) {
		WQ_WAIT_F(&geth_wait, rq);
		return NULL;
	}
	memcpy(&geth_vdesc, desc, sizeof(VDESC));
	geth_caller = rq;
	geth_wr = wr;
	geth.tmp1 = (unsigned long)NOOP_CONT_PAGEIN;
	return &geth;
}

DECL_IOCALL(NOOP_CONT_PAGEIN, SPL_DEV, IORQ)
{
	IORQ *caller;
	HANDLE *h = geth_vdesc.vspace;
	if (__unlikely(h->op != &KERNEL_HANDLE_NOOP)) goto x;
	if (__unlikely(KERNEL$LOCKUP_LEVEL >= LOCKUP_LEVEL_ALL_IORQS)) {
		WQ_WAIT_F(&KERNEL$LOCKUP_EVENTS, geth_caller);
		goto free_geth;
	}
	/* SWITCH_PROC_ACCOUNT(h->name_addrspace, SPL_X(SPL_DEV)); will be done either by GET_HANDLE or RETURN_IORQ_LSTAT */
	h = GET_HANDLE(h->name_addrspace, h->handle_num & HANDLE_NUM_MASK, 0, geth_caller);
	if (!h) {
		free_geth:
		RAISE_SPL(SPL_VSPACE);
		geth_caller = NULL;
		WQ_WAKE_ALL(&geth_wait);
		LOWER_SPL(SPL_DEV);
		RETURN;
	}
	if (__unlikely(__IS_ERR(h))) {
		caller = geth_caller;
		caller->status = __PTR_ERR(h);
		RAISE_SPL(SPL_VSPACE);
		geth_caller = NULL;
		WQ_WAKE_ALL(&geth_wait);
		LOWER_SPL(SPL_DEV);
		RETURN_AST(caller);
	}
	caller = geth_caller;
	RAISE_SPL(SPL_VSPACE);
	geth_caller = NULL;
	WQ_WAKE_ALL(&geth_wait);
	LOWER_SPL(SPL_DEV);
	RETURN_IORQ_LSTAT(caller, (IO_STUB *)caller->tmp1);

	x:
	RAISE_SPL(SPL_VSPACE);
	caller = geth_caller;
	geth_caller = NULL;
	WQ_WAKE_ALL(&geth_wait);
	DO_PAGEIN(caller, &geth_vdesc, geth_wr);
}

typedef struct {
	__PACKET_HANDLE_ENTRIES;
} PKT_HEADER;

static void RET_PKT(PKT_HEADER *RQ)
{
	void (*call_pkt)(PKT_HEADER *);
	call_pkt = dlsym(RTLD_DEFAULT, "NET$CALL_PKT");
	if (__unlikely(!call_pkt))
		KERNEL$SUICIDE("RET_PKT: PACKET ENCOUNTERED AND NET.DLL NOT LINKED");
	call_pkt(RQ);
}

#define NOP_OP(name, opp, rq, call)					\
DECL_IOCALL(name, SPL_DEV, rq)						\
{									\
	HANDLE *h = RQ->handle;						\
	if (__unlikely(h->op != &KERNEL_HANDLE_NOOP)) goto x;		\
	RQ->tmp1 = (unsigned long)name;					\
	TEST_LOCKUP_ENTRY(RQ, RETURN);					\
	/* SWITCH_PROC_ACCOUNT(h->name_addrspace, SPL_X(SPL_DEV)); will be done either by GET_HANDLE or RETURN_IORQ_LSTAT */				\
	h = GET_HANDLE(h->name_addrspace, h->handle_num & HANDLE_NUM_MASK, 0, (IORQ *)RQ);								\
	if (!h) RETURN;							\
	if (__unlikely(__IS_ERR(h))) {					\
		RQ->status = __PTR_ERR(h);				\
		call(RQ);						\
		RETURN;							\
	}								\
	x:								\
	RETURN_IORQ_LSTAT(RQ, opp);					\
}

NOP_OP(NOOP_READ, KERNEL$WAKE_READ, SIORQ, RETURN_AST)
NOP_OP(NOOP_WRITE, KERNEL$WAKE_WRITE, SIORQ, RETURN_AST)
NOP_OP(NOOP_AREAD, KERNEL$WAKE_AREAD, AIORQ, RETURN_AST)
NOP_OP(NOOP_AWRITE, KERNEL$WAKE_AWRITE, AIORQ, RETURN_AST)
NOP_OP(NOOP_IOCTL, KERNEL$WAKE_IOCTL, IOCTLRQ, RETURN_AST)
NOP_OP(NOOP_BIO, KERNEL$WAKE_BIO, BIORQ, RETURN_AST)
NOP_OP(NOOP_PKTIO, KERNEL$WAKE_PKTIO, PKT_HEADER, RET_PKT)

const HANDLE_OPERATIONS KERNEL_HANDLE_NOOP = {
	SPL_X(SPL_DEV),
	KERNEL$NO_VSPACE_GET,
	KERNEL$NO_VSPACE_PUT,
	KERNEL$NO_VSPACE_MAP,
	KERNEL$NO_VSPACE_DMALOCK,
	KERNEL$NO_VSPACE_DMA64LOCK,
	KERNEL$NO_VSPACE_PHYSLOCK,
	NOOP_GET_PAGEIN_RQ,
	KERNEL$NO_VSPACE_GET_PAGE,
	KERNEL$NO_VSPACE_SWAP_OP,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NOOP_READ,
	NOOP_WRITE,
	NOOP_AREAD,
	NOOP_AWRITE,
	NOOP_IOCTL,
	NOOP_BIO,
	NOOP_PKTIO
};

const HANDLE_OPERATIONS empty_handle = {
	SPL_X(SPL_DEV),
	KERNEL$NO_VSPACE_GET,
	KERNEL$NO_VSPACE_PUT,
	KERNEL$NO_VSPACE_MAP,
	KERNEL$NO_VSPACE_DMALOCK,
	KERNEL$NO_VSPACE_DMA64LOCK,
	KERNEL$NO_VSPACE_PHYSLOCK,
	KERNEL$NO_VSPACE_GET_PAGEIN_RQ,
	KERNEL$NO_VSPACE_GET_PAGE,
	KERNEL$NO_VSPACE_SWAP_OP,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	KERNEL$NO_OPERATION,
	KERNEL$NO_OPERATION,
	KERNEL$NO_OPERATION,
	KERNEL$NO_OPERATION,
	KERNEL$NO_OPERATION,

	KERNEL$NO_OPERATION,
	KERNEL$NO_OPERATION,
};

extern AST_STUB bio_emu_endrq;
extern AST_STUB bio_emu_endflush;

DECL_IOCALL(KERNEL$BIO_EMU, SPL_FS, BIORQ)
{
	_u_off_t o;
	BIODESC *desc;
	SWITCH_PROC_ACCOUNT(RQ->proc, SPL_X(SPL_FS));
	BIO_ASSERT_FAULT_SEC;
	if (__likely(!(RQ->flags & (BIO_WRITE | BIO_FLUSH)))) {
		RQ->tmp1 = (unsigned long)KERNEL$AREAD;
		goto io;
	} else if (__likely(RQ->flags & BIO_WRITE)) {
		RQ->tmp1 = (unsigned long)KERNEL$AWRITE;
		goto io;
	} else if (__likely(RQ->flags & BIO_FLUSH)) {
#define IRQ	(&((_IBIORQ *)RQ)->i)
		IRQ->h = RQ->h;
		__barrier();
		IRQ->fn = bio_emu_endflush;
		IRQ->ioctl = IOCTL_FSYNC;
		IRQ->param = PARAM_FSYNC_ONLY_DATA;
		IRQ->v.ptr = 0;
		IRQ->v.len = 0;
		IRQ->v.vspace = &KERNEL$VIRTUAL;
		RETURN_IORQ_CANCELABLE(IRQ, KERNEL$IOCTL, RQ);
#undef IRQ
	} else KERNEL$SUICIDE("KERNEL$BIO_EMU: INVALID FLAGS %X", RQ->flags);
	io:
	/*__debug_printf("BIO EMU: %p->(%d)%p\n", RQ, __offsetof(BIORQ, proc), RQ->proc);*/
	if (__unlikely(RQ->sec > (_u_off_t)-1 >> BIO_SECTOR_SIZE_BITS)) {
		RQ->status = -ERANGE;
		RETURN_AST(RQ);
	}
	o = (_u_off_t)RQ->sec << BIO_SECTOR_SIZE_BITS;
#define ARQ	(&((_ABIORQ *)RQ)->a)
	desc = RQ->desc;
	/* handling of page-faults is disabled in KERNEL$PAGEIN if AST is bio_emu_endrq */
	RQ->tmp3 = (unsigned long)desc->next;
	ARQ->h = RQ->h;
	__barrier();
	ARQ->fn = bio_emu_endrq;
	ARQ->v.ptr = desc->v.ptr;
	ARQ->v.len = desc->v.len;
	ARQ->v.vspace = desc->v.vspace;
	ARQ->pos = o;
	ARQ->progress = 0;
	RETURN_IORQ_CANCELABLE(ARQ, (IO_STUB *)RQ->tmp1, RQ);
#undef ARQ
}

DECL_AST(bio_emu_endrq, SPL_FS, AIORQ)
{
	_ABIORQ *a = GET_STRUCT(RQ, _ABIORQ, a);
	BIODESC *desc;
#define b	 ((BIORQ *)a)
	IO_DISABLE_CHAIN_CANCEL(SPL_FS, a);
	/*__debug_printf("BIO EMU ENDRQ: %p->(%d)%p\n", RQ, __offsetof(BIORQ, proc), RQ->proc);*/
	if (__unlikely(a->a.status <= 0)) {
		int h;
		b->status = __likely(a->a.status != 0) ? a->a.status : -ERANGE;
		h = a->a.h;
		b->fault_sec = a->a.pos >> BIO_SECTOR_SIZE_BITS;
		__barrier();
		b->h = h;
		RETURN_BIO(b);
	}
	if (__unlikely(a->a.v.len != 0)) {
		a->a.progress = 0;
		RETURN_IORQ_CANCELABLE(&a->a, (IO_STUB *)b->tmp1, b);
	}
	desc = (BIODESC *)a->tmp3;
	if (__likely(!desc)) {
		b->status = 0;
		b->h = a->a.h;
		RETURN_BIO(b);
	}
	b->tmp3 = (unsigned long)desc->next;
	a->a.v.ptr = desc->v.ptr;
	a->a.v.len = desc->v.len;
	a->a.v.vspace = desc->v.vspace;
	a->a.progress = 0;
	RETURN_IORQ_CANCELABLE(&a->a, (IO_STUB *)b->tmp1, b);
#undef b
}

DECL_AST(bio_emu_endflush, SPL_FS, IOCTLRQ)
{
	BIORQ *b = (BIORQ *)GET_STRUCT(RQ, _IBIORQ, i);
	IO_DISABLE_CHAIN_CANCEL(SPL_FS, b);
	b->status = ((_IBIORQ *)b)->i.status;
	__barrier();
	b->fault_sec = -1;
	RETURN_BIO(b);
}

int substitute_option(char *src, char *dest, char *option, char *value)
{
	char *p;
	p = strchr(src, ':');
	if (__unlikely(!p)) return -EBADSYN;
	p++;
	if (__unlikely(!*p)) {
		e:
		if (value) {
			strcpy(dest, src);
			if (dest[strlen(dest) - 1] == '/') strcat(dest, "^");
			else strcat(dest, "/^");
			strcat(dest, option);
			if (*value) {
				strcat(dest, "=");
				strcat(dest, value);
			}
			return 1;
		} else return 0;
	}
	f:
	if (*p == '^') {
		char *o = option, *op = p + 1;
		while (*o && __upcasechr(*o) == __upcasechr(*op)) o++, op++;
		if (*o || (*op && *op != '=' && *op != '/')) {
			p = op;
			goto d;
		}
		if (value) {
			char *oo, *oop;
			if (!*op || *op == '/') {
				if (!*value) return 0;
				memcpy(dest, src, op - src);
				strcpy(dest + (op - src), "=");
				strcat(dest + (op - src), value);
				strcat(dest + (op - src), op);
				return 1;
			}
			op++;
			oo = value, oop = op;
			while (*oo && __upcasechr(*oo) == __upcasechr(*oop)) oo++, oop++;
			if (!*oo && (!*oop || *oop == '/')) return 0;
			while (*oop && *oop != '/') oop++;
			if (!*value) op--;
			memcpy(dest, src, op - src);
			strcpy(dest + (op - src), value);
			strcat(dest + (op - src), oop);
			return 1;
		} else {
			memcpy(dest, src, p - src);
			while (*op && *op != '/') op++;
			if (dest[p - src - 1] == '/' && *op == '/') op++;
			strcpy(dest + (p - src), op);
			return 1;
		}
	}
	d:
	p = strchr(p, '/');
	if (!p) goto e;
	p++;
	goto f;
}

#define split	put_ioctl_split
#define fn	KERNEL$PUT_IOCTL_STRUCT
#define put	1

#include "GETPUTIO.I"

#define split	get_ioctl_split
#define fn	KERNEL$GET_IOCTL_STRUCT
#define put	0

#include "GETPUTIO.I"


int LN_TEST_RECURSIVE_DEPENDENCIES(LNTE *lnte, int (*init)(LNTE *, void *), int (*test)(LNTE *, void *), void *p)
{
	LIST_HEAD deps;
	unsigned i;
	LNTE *original = lnte;
	LNTE *e;

#if __DEBUG >= 1
	if (__unlikely(KERNEL$SPL != SPL_X(SPL_DEV)))
		KERNEL$SUICIDE("LN_TEST_RECURSIVE_DEPENDENCIES AT SPL %08X", KERNEL$SPL);
#endif

	INIT_LIST(&deps);
	for (i = 0; i < LNT_HASH_SIZE; i++) XLIST_FOR_EACH(e, &LN->hash[i], LNTE, hash) {
		if (init && __unlikely(init(e, p)))
			ADD_TO_LIST(&deps, &e->depwalk);
		else
			e->depwalk.next = NULL;
	}

	if (!lnte) goto get_dep;

	process_dep:
	for (i = 0; i < lnte->n_deps; i++) {
		char *n = lnte->deps[i];
		char *c = n;
		int hash = 0;
		if (*c == ':') continue;
		quickcasehash(c, *c != ':', hash);
		hash &= LNT_HASH_SIZE - 1;

		if (original && __unlikely(!__strcasexcmp(original->name, n, c)))
			return -ELOOP;

		XLIST_FOR_EACH(e, &LN->hash[hash], LNTE, hash) if (__likely(!__strcasexcmp(e->name, n, c))) {
			if (!e->depwalk.next) {
				ADD_TO_LIST_END(&deps, &e->depwalk);
			}
			break;
		}
	}

	get_dep:
	if (!LIST_EMPTY(&deps)) {
		int r;
		lnte = LIST_STRUCT(deps.next, LNTE, depwalk);
		DEL_FROM_LIST(&lnte->depwalk);
		lnte->depwalk.next = (void *)1;
		if (test && __unlikely(r = test(lnte, p)))
			return r;
		goto process_dep;
	}

	return 0;
}

int KERNEL$GET_DEPENDENCY(const char *name, unsigned n, char *ret, int ret_size)
{
	const char *c;
	int r;
	int hash;
	LNTE *e;
	int spl = KERNEL$SPL;
	RAISE_SPL(SPL_DEV);
	c = name;
	hash = 0;
	quickcasehash(c, *c && *c != ':', hash);
	hash &= LNT_HASH_SIZE - 1;
	XLIST_FOR_EACH(e, &LN->hash[hash], LNTE, hash) if (__likely(!__strcasexcmp(e->name, name, c))) {
		if (n >= e->n_deps) {
			r = -ENOENT;
			goto ret;
		}
		r = strlcpy(ret, e->deps[n], ret_size) + 1;
		goto ret;
	}
	r = -ENOLNM;
	ret:
	LOWER_SPLX(spl);
	return r;
}

int LN_INIT_VM(LNTE *lnte, void *p)
{
	return __likely(lnte != p) && (lnte->handle & (LNTE_DEVICE | LNTE_VM_ESSENTIAL)) == (LNTE_DEVICE | LNTE_VM_ESSENTIAL);
}

int LN_TEST_EQ(LNTE *lnte, void *p)
{
	return __unlikely(lnte == p) ? -EBUSY : 0;
}

int KERNEL$DELETE_LOGICAL(char *lnm, int flags)
{
	int r;
	int spl;
	int h;
	LNTE *e;
	char *c;
	int handle_to_close;
#if __DEBUG >= 1
	if (__unlikely(flags & ~(LN_DELETE_LOGICAL_BLOCK | LN_DONT_WAKE_ALL)))
		KERNEL$SUICIDE("KERNEL$DELETE_LOGICAL: INVALID FLAGS: %X (LNM %s)", flags, lnm);
#endif
	if (SPLX_BELOW(SPL_X(SPL_DEV), spl = KERNEL$SPL)) KERNEL$SUICIDE("KERNEL$DELETE_LOGICAL AT SPL %08X", KERNEL$SPL);
	RAISE_SPL(SPL_DEV);

	if (__check_logical_name(lnm, 0)) {
		r = -EBADSYN;
		goto ret;
	}
	c = lnm;
	h = 0;
	quickcasehash(c, *c && *c != ':', h);
	h &= LNT_HASH_SIZE - 1;

	XLIST_FOR_EACH(e, &LN->hash[h], LNTE, hash) if (__likely(!__strcasexcmp(e->name, lnm, c))) goto found;
	r = -ENOLNM;
	goto ret;

	found:
	if (e->handle & LNTE_DEVICE) {
		r = -EBUSY;
		goto ret;
	}
	if (__unlikely(r = LN_TEST_RECURSIVE_DEPENDENCIES(NULL, LN_INIT_VM, LN_TEST_EQ, e)))
		goto ret;
	handle_to_close = e->handle & LNTE_HANDLE_MASK;
	if (flags & LN_DELETE_LOGICAL_BLOCK) {
		e->handle = (e->handle & (LNTE_PUBLIC | LNTE_FILE_SHARE)) | BLOCK_HANDLE;
	} else {
		DEL_FROM_LIST_ORDERED(&e->hash);
		KERNEL$UNIVERSAL_FREE(e);
	}
	MULTIOPEN_CLEAR_CACHE();
	if (__likely(!(flags & LN_DONT_WAKE_ALL)))
		SWAP_ALL_SUBPROCESSES(); /* page mappings are not bound on handles --- need to erase them */
	CLEAR_SUBPROC_LN_CACHE();
	if (__likely(handle_to_close != BLOCK_HANDLE)) KERNEL$FAST_CLOSE(handle_to_close);
	else UNBLOCK_LNT();
	WQ_WAKE_ALL(&KERNEL$LOGICAL_WAIT);
	/* DO_BREAK(&KERNEL$PROC_KERNEL, NULL, 1); done in SWAP_ALL_SUBPROCESSES */
	r = 0;

	ret:
	LOWER_SPLX(spl);
	return r;
}

int KERNEL$CREATE_LOGICAL(const char *name, const char *alias, int flags)
{
	int h, r, hash;
	int spl;
	LNTE *lnte, *e;
	char *c;
	char *x, *hp;
	union {
		OPENRQ openrq;
		MALLOC_REQUEST mrq;
	} u;
#if __DEBUG >= 1
	if (__unlikely(flags & ~(LNTE_PUBLIC | LNTE_BLOCK_IF_NOT_FOUND | LNTE_FILE_SHARE | LN_DONT_WAKE_ALL)))
		KERNEL$SUICIDE("KERNEL$CREATE_LOGICAL: INVALID FLAGS: %X (LNM %s->%s)", flags, alias, name);
#endif
	spl = KERNEL$SPL;
	RAISE_SPL(SPL_DEV);
	u.openrq.flags = (__unlikely(flags & LNTE_FILE_SHARE) ? O_RDWR : _O_NOACCESS) | _O_NOOPEN | _O_NOPOSIX;
	u.openrq.path = alias;
	u.openrq.cwd = KERNEL$CWD();
	open_again:
	SYNC_IO_CANCELABLE(&u.openrq, KERNEL$OPEN);
	if (__unlikely(u.openrq.status < 0)) {
		if (__likely((u.openrq.flags & 7) != _O_NOACCESS)) {
			if (__likely((u.openrq.flags & 7) != O_RDONLY))
				u.openrq.flags--;
			else u.openrq.flags = (u.openrq.flags & ~7) | _O_NOACCESS;
			goto open_again;
		}
		r = u.openrq.status;
		goto ret0;
	}
	h = u.openrq.status;
	hp = KERNEL$HANDLE_PATH(h);

	u.mrq.size = sizeof(LNTE) + strlen(name) + 2 * sizeof(char *) - 1 + strlen(hp) + 1;
	SYNC_IO_CANCELABLE(&u.mrq, KERNEL$UNIVERSAL_MALLOC);
	if (__unlikely(u.mrq.status < 0)) {
		r = u.mrq.status;
		goto ret1;
	}
	lnte = u.mrq.ptr;
	lnte->handle = h | (flags & (LNTE_PUBLIC | LNTE_BLOCK_IF_NOT_FOUND | LNTE_FILE_SHARE));
	lnte->dcall = NULL;
	lnte->dcall_type = NULL;
	lnte->dctl = NULL;
	lnte->unload = NULL;
	lnte->driver_name = NULL;
	x = stpcpy(lnte->name, name) + 1;
	x = (char *)(((unsigned long)x + (sizeof(char *) - 1)) & ~(unsigned long)(sizeof(char *) - 1));
	*(lnte->deps = (char **)x) = x + sizeof(char *);
	x += sizeof(char *);
	strcpy(x, hp);
	lnte->n_deps = 1;
	if (__unlikely(__check_logical_name(lnte->name, 1))) {
		r = -EINVAL;
		goto ret1;
	}
	c = lnte->name;
	hash = 0;
	quickcasehash(c, *c, hash);
	hash &= LNT_HASH_SIZE - 1;

	XLIST_FOR_EACH(e, &LN->hash[hash], LNTE, hash) if (__likely(!__strcasexcmp(e->name, lnte->name, c))) {
		char *h1, *h2;
		if (__unlikely(e->handle & LNTE_DEVICE)) goto do_delete;
		h1 = KERNEL$HANDLE_PATH(lnte->handle & LNTE_HANDLE_MASK);
		h2 = KERNEL$HANDLE_PATH(e->handle & LNTE_HANDLE_MASK);
		if (__unlikely(!h1) || __unlikely(!h2)) goto do_delete;
		if (strcmp(h1, h2)) goto do_delete;
		if (__unlikely((lnte->handle & ~LNTE_HANDLE_MASK) != (e->handle & ~LNTE_HANDLE_MASK))) goto do_delete;
		/* !!! FIXME: test blob */
		r = 0;
		goto ret2;

		do_delete:
		if (__unlikely(r = KERNEL$DELETE_LOGICAL(e->name, 0))) {
			if (r == -EBUSY) r = -EEXIST;
			goto ret2;
		}
		break;
	}
	if (__unlikely(r = LN_TEST_RECURSIVE_DEPENDENCIES(lnte, NULL, NULL, NULL)))
		goto ret2;
	ADD_TO_XLIST_ORDERED(&LN->hash[hash], &lnte->hash);
	WQ_WAKE_ALL(&KERNEL$LOGICAL_WAIT);
	MULTIOPEN_CLEAR_CACHE();
	UNBLOCK_LNT();
	if (__likely(!(flags & LN_DONT_WAKE_ALL))) {
		if (__likely(!KERNEL$KERNEL)) {
			USR_SWAP_ALL_HANDLES();
			SWAP_ALL_SUBPROCESSES();
		} else DO_BREAK(&KERNEL$PROC_KERNEL, NULL, 1);
	}
	CLEAR_SUBPROC_LN_CACHE();
	r = 0;
	goto ret0;

	ret2:
	KERNEL$UNIVERSAL_FREE(lnte);
	ret1:
	KERNEL$FAST_CLOSE(h);
	ret0:
	LOWER_SPLX(spl);
	return r;
}

int KERNEL$SET_LOGICAL(const char *name, int flags, int flags_mask, char *chh_option, char *chh_value, FBLOB *blob)
{
	int need_swap = 0;
	int newh, h, r, hash;
	int spl;
	LNTE *e;
	const char *c;
	char *tmp_alias;
	union {
		CHHRQ chh;
		HBLOBRQ brq;
		MALLOC_REQUEST mrq;
	} u;
#if __DEBUG >= 1
	if (__unlikely(flags & ~(LNTE_PUBLIC | LNTE_BLOCK_IF_NOT_FOUND | LNTE_FILE_SHARE | LN_DONT_WAKE_ALL)) || __unlikely(flags_mask & ~(LNTE_PUBLIC | LNTE_BLOCK_IF_NOT_FOUND | LNTE_FILE_SHARE)))
		KERNEL$SUICIDE("KERNEL$SET_LOGICAL: INVALID FLAGS: %X/%X (LNM %s)", flags, flags_mask, name);
#endif
	spl = KERNEL$SPL;
	RAISE_SPL(SPL_DEV);

	again:
	c = name;
	if (__check_logical_name((char *)c, 0)) {
		r = -EINVAL;
		goto ret0;
	}
	hash = 0;
	quickcasehash(c, *c && *c != ':', hash);
	hash &= LNT_HASH_SIZE - 1;

	XLIST_FOR_EACH(e, &LN->hash[hash], LNTE, hash) if (__likely(!__strcasexcmp(e->name, name, c))) goto found;
	if (KERNEL$KERNEL) {
		r = -ENOLNM;
		goto ret0;
	}
	u.mrq.size = c - name + 4;
	SYNC_IO_CANCELABLE(&u.mrq, KERNEL$UNIVERSAL_MALLOC);
	if (__unlikely(u.mrq.status < 0)) {
		r = u.mrq.status;
		goto ret0;
	}
	tmp_alias = u.mrq.ptr;
	stpcpy(mempcpy(stpcpy(tmp_alias, "@"), name, c - name), ":/");
		/* !!! TODO: query public state, if not chaging it */
	r = KERNEL$CREATE_LOGICAL(name, tmp_alias, flags);
	KERNEL$UNIVERSAL_FREE(tmp_alias);
	if (__unlikely(r < 0)) {
		goto ret0;
	}
	goto again;

	found:
	if (__unlikely(e->handle & LNTE_DEVICE) && __unlikely(((unsigned long)chh_option | (unsigned long)blob) != 0)) {
		r = -EOPNOTSUPP;
		goto ret0;
	}
	newh = (e->handle & (~flags_mask | LNTE_HANDLE_MASK | LNTE_DEVICE)) | (flags & (LNTE_PUBLIC | LNTE_BLOCK_IF_NOT_FOUND | LNTE_FILE_SHARE));
	if ((unsigned long)chh_option | (unsigned long)blob) {
		newh = (newh & ~LNTE_PUBLIC) | (e->handle & newh & LNTE_PUBLIC);
	}
	if (newh != e->handle || need_swap) {
		e->handle = newh;
		WQ_WAKE_ALL(&KERNEL$LOGICAL_WAIT);
		MULTIOPEN_CLEAR_CACHE();
		if (__likely(!(flags & LN_DONT_WAKE_ALL))) {
			USR_SWAP_ALL_HANDLES();
			SWAP_ALL_SUBPROCESSES();
		}
		CLEAR_SUBPROC_LN_CACHE();
		need_swap = 0;
	}
	if (__unlikely(((unsigned long)chh_option | (unsigned long)blob) != 0)) {
		h = _dup_noposix(newh & LNTE_HANDLE_MASK, NULL, NULL);
		if (__unlikely(h < 0)) {
			if (h == -EAGAIN) goto again;
			r = h;
			goto ret0;
		}
		if (chh_option) {
			u.chh.h = h;
			u.chh.option = chh_option;
			u.chh.value = chh_value;
			chh_option = NULL;
			SYNC_IO_CANCELABLE(&u.chh, KERNEL$CHANGE_HANDLE);
			if (__unlikely(u.chh.status < 0)) {
				r = u.chh.status;
				goto ret1;
			}
			need_swap = 1;
		}
		if (blob) {
			u.brq.h = h;
			u.brq.blob = blob;
			blob = NULL;
			SYNC_IO_CANCELABLE(&u.brq, KERNEL$SET_HANDLE_BLOB);
			if (__unlikely(u.brq.status < 0)) {
				r = u.brq.status;
				goto ret1;
			}
			need_swap = 1;
		}
		KERNEL$FAST_CLOSE(h);
		goto again;
	}
	r = 0;
	goto ret0;

	ret1:
	KERNEL$FAST_CLOSE(h);
	ret0:
	LOWER_SPLX(spl);
	return r;
}

static int CMP_LOGICAL_ENTRY(const void *v1, const void *v2)
{
	const LOGICAL_LIST_ENTRY *e1 = *(LOGICAL_LIST_ENTRY **)v1;
	const LOGICAL_LIST_ENTRY *e2 = *(LOGICAL_LIST_ENTRY **)v2;
	return strcmp(e1->name, e2->name);
}

void SORT_LOGICAL_LIST(LOGICAL_LIST_REQUEST *r)
{
	int i, j, sub;
	qsort(r->entries, r->n_entries, sizeof(LOGICAL_LIST_ENTRY *), CMP_LOGICAL_ENTRY);
	for (i = r->n_entries - 1; i > 0; i--) {
		if (__unlikely(!strcmp(r->entries[i - 1]->name, r->entries[i]->name))) goto shift;
	}
	shift_done:
	j = 0;
	for (i = 0; i < r->n_entries; i++) {
		/*if (__likely(r->entries[i]->flags != LNTE_INVALID))*/ r->entries[j++] = r->entries[i];
	}
	r->n_entries = j;
	return;
	shift:
	sub = 0;
	j = 1;
	for (i = 1; i < r->n_entries; i++) {
		if (__likely(strcmp(r->entries[j - 1]->name, r->entries[i]->name))) {
			r->entries[j++] = r->entries[i];
		} else {
			if (r->entries[j - 1]->flags == LNTE_PARENT /*&& r->entries[i]->flags != LNTE_PARENT*/) r->entries[j - 1] = r->entries[i];
			sub++;
		}
	}
	r->n_entries -= sub;
	goto shift_done;
}

void KERNEL$FREE_LOGICAL_LIST(LOGICAL_LIST_REQUEST *r)
{
	while (!XLIST_EMPTY(&r->freelist)) {
		LOGICAL_LIST_ENTRY *e = LIST_STRUCT(r->freelist.next, LOGICAL_LIST_ENTRY, list);
		DEL_FROM_LIST(&e->list);
		free(e);
	}
	free(r->entries);
}

struct multiopen {
	OPENRQ *rq;
	FFILE *f;
	char *ne;
	int llr_valid;
	time_t llr_time;
	int idx;
	int error;
	int error_pri;
	LOGICAL_LIST_REQUEST llr;
	OPENRQ o;
	MALLOC_REQUEST mrq;
};

extern struct multiopen *multiopen;

extern AST_STUB MULTIOPEN_GOT_LIST;
extern IO_STUB MULTIOPEN_TRY_OPEN;
extern AST_STUB MULTIOPEN_NAME_ALLOCATED;
extern AST_STUB MULTIOPEN_OPEN;

void MULTIOPEN_CLEAR_CACHE(void)
{
	if (__unlikely(KERNEL$SPL != SPL_X(SPL_DEV)))
		KERNEL$SUICIDE("MULTIOPEN_CLEAR_CACHE AT SPL %08X", KERNEL$SPL);
	if (__likely(multiopen != NULL)) multiopen->llr_time = -1;
}

int MULTIOPEN_ERROR_PRI(int error)
{
#if defined(__GNUC__) && __GNUC__ >= 3
	__builtin_expect(error, -ENOENT);
#endif
	switch (error) {
		case -ENOLNM:	return 1;
		case -ENOENT:	return 2;
		case -EACCES:
		case -EPERM:
		case -EROFS:	return 3;
		case -EISDIR:
		case -ENOTDIR:	return 4;
		default:	return 5;
	}
}

static void MULTIOPEN_FREE(struct multiopen *mo)
{
	free(mo->f);
	if (__likely(!multiopen)) {
		multiopen = mo;
		return;
	}
	if (__likely(mo->llr_valid) && (__unlikely(!multiopen->llr_valid) || __unlikely((unsigned)multiopen->llr_time < (unsigned)mo->llr_time))) {
		struct multiopen *mo_x = multiopen;
		multiopen = mo;
		mo = mo_x;
	}
	if (__likely(mo->llr_valid)) KERNEL$FREE_LOGICAL_LIST(&mo->llr);
	free(mo);
}

void MULTIOPEN(OPENRQ *RQ, FFILE *f)
{
	struct multiopen *mo;
	if (__unlikely(!multiopen)) {
		if (__unlikely(!(mo = malloc(sizeof(struct multiopen))))) {
			__slow_free(f);
			KERNEL$MEMWAIT((IORQ *)RQ, KERNEL$OPEN, sizeof(struct multiopen));
			return;
		}
		mo->llr_valid = 0;
	} else {
		mo = multiopen;
		multiopen = NULL;
	}
	mo->rq = RQ;
	mo->f = f;
	mo->ne = strchr(f->path, ':');
	if (__unlikely(!mo->ne) || __unlikely(mo->ne == f->path)) KERNEL$SUICIDE("MULTIOPEN: PATH DOESN'T CONTAIN COLON: \"%s\" -> \"%s\"", RQ->path, f->path);
	mo->ne--;
	if (__unlikely(mo->ne[0] != '.')) {
		RQ->status = -EINVAL;
		MULTIOPEN_FREE(mo);
		CALL_AST(RQ);
		return;
	}
	mo->idx = -1;
	if (__unlikely(!mo->llr_valid)) {
		list:
		mo->llr.prefix = "";	/* !!! TODO: use prefixes ? */
		mo->llr.fn = MULTIOPEN_GOT_LIST;
		CALL_IORQ_CANCELABLE(&mo->llr, KERNEL$LIST_LOGICALS, RQ);
		return;
	}
	if (__unlikely(mo->llr_time != time(NULL))) {
		KERNEL$FREE_LOGICAL_LIST(&mo->llr);
		mo->llr_valid = 0;
		goto list;
	}
	CALL_IORQ(&mo->o, MULTIOPEN_TRY_OPEN);
	return;
}

DECL_AST(MULTIOPEN_GOT_LIST, SPL_DEV, LOGICAL_LIST_REQUEST)
{
	OPENRQ *rq;
	struct multiopen *mo = LIST_STRUCT(RQ, struct multiopen, llr);
	IO_DISABLE_CHAIN_CANCEL(SPL_DEV, mo->rq);
	if (__unlikely(RQ->status < 0)) {
		rq = mo->rq;
		rq->status = RQ->status;
		MULTIOPEN_FREE(mo);
		RETURN_AST(rq);
	}
	mo->llr_valid = 1;
	time(&mo->llr_time);
	RETURN_IORQ(&mo->o, MULTIOPEN_TRY_OPEN);
}

DECL_IOCALL(MULTIOPEN_TRY_OPEN, SPL_DEV, OPENRQ)
{
	OPENRQ *rq;
	struct multiopen *mo = LIST_STRUCT(RQ, struct multiopen, o);
	if (mo->idx == -1) {
		int i;
		mo->error_pri = MULTIOPEN_ERROR_PRI(mo->error = -ENOLNM);
		for (i = 0; i < mo->llr.n_entries; i++) if (__unlikely(!__memcasexcmp(mo->llr.entries[i]->name, mo->f->path, mo->ne))) {
			mo->idx = i;
			goto try_it;
		}
		err:
		rq = mo->rq;
		rq->status = mo->error;
		MULTIOPEN_FREE(mo);
		RETURN_AST(rq);
	}
	if (__unlikely(mo->idx >= mo->llr.n_entries) || __memcasexcmp(mo->llr.entries[mo->idx]->name, mo->f->path, mo->ne)) {
		goto err;
	}
	try_it:
	mo->mrq.size = strlen(mo->llr.entries[mo->idx]->name) + strlen(mo->ne);
	mo->mrq.fn = MULTIOPEN_NAME_ALLOCATED;
	RETURN_IORQ_CANCELABLE(&mo->mrq, KERNEL$UNIVERSAL_MALLOC, mo->rq);
}

DECL_AST(MULTIOPEN_NAME_ALLOCATED, SPL_DEV, MALLOC_REQUEST)
{
	char *p;
	OPENRQ *rq;
	struct multiopen *mo = LIST_STRUCT(RQ, struct multiopen, mrq);
	IO_DISABLE_CHAIN_CANCEL(SPL_DEV, mo->rq);
	if (__unlikely(RQ->status < 0)) {
		rq = mo->rq;
		rq->status = RQ->status;
		MULTIOPEN_FREE(mo);
		RETURN_AST(rq);
	}
	p = RQ->ptr;
	stpcpy(stpcpy(p, mo->llr.entries[mo->idx]->name), mo->ne + 1);
	mo->o.flags = mo->rq->flags;
	if (__unlikely(mo->o.flags & _O_NOOPEN)) {
		mo->o.flags = (mo->o.flags & ~_O_NOOPEN) | _O_NOOPEN_CALL;
	}
	mo->o.path = p;
	mo->o.rename_handle = mo->rq->rename_handle;
	mo->o.cwd = NULL;
	mo->o.fn = MULTIOPEN_OPEN;
	RETURN_IORQ_CANCELABLE(&mo->o, KERNEL$OPEN, mo->rq);
}

DECL_AST(MULTIOPEN_OPEN, SPL_DEV, OPENRQ)
{
	OPENRQ *rq;
	struct multiopen *mo = LIST_STRUCT(RQ, struct multiopen, o);
	IO_DISABLE_CHAIN_CANCEL(SPL_DEV, mo->rq);
	KERNEL$UNIVERSAL_FREE(mo->mrq.ptr);
	if (RQ->status < 0) {
		int pri;
		if (__unlikely(RQ->status == -EINTR)) {
			got_it:
			rq = mo->rq;
			rq->status = RQ->status;
			MULTIOPEN_FREE(mo);
			RETURN_AST(rq);
		}
		pri = MULTIOPEN_ERROR_PRI(RQ->status);
		if (pri > mo->error_pri) {
			mo->error = RQ->status;
			mo->error_pri = pri;
		}
		mo->idx++;
		RETURN_IORQ_LSTAT(&mo->o, MULTIOPEN_TRY_OPEN);
	}
	goto got_it;
}

int is_exec(const char *orig_p)
{
	char *p = (char *)orig_p;
	int r, i;
	struct link_header lh;
#define lhc ((char *)&lh)
	int h;
	int rcount = 0;
	reopen:
	h = open(p, O_RDONLY | _O_NOPOSIX);
	if (__unlikely(h == -1)) {
		if (p != orig_p) errno = ENOEXEC;
		return 0;
	}
	r = read(h, &lh, sizeof lh);
	if (__likely(r > 2) && __unlikely(lhc[0] == '#') && __unlikely(lhc[1] == '!') && __likely(p == orig_p)) {
		close(h);
		return 1;
	}
	if (__unlikely(r != sizeof lh)) {
		if (__unlikely(r == -1)) {
			close(h);
			return 0;
		}
		if (__unlikely(r < 2) || __unlikely(memchr(lhc, '\n', r) != &lhc[r - 1]) || __unlikely(memchr(lhc, 0, r) != 0)) goto c_enoexec;
		if (__unlikely(++rcount > LINK_MAX_RCOUNT)) {
			close(h);
			errno = ELOOP;
			return 0;
		}
		r--;
		lhc[r] = 0;
		i = _is_absolute(lhc);
		if (__unlikely(i == _ABS_LOGREL) || __unlikely(i == _ABS_CURROOT)) goto c_enoexec;
		if (__unlikely(i == _ABS_TOTAL)) {
			close(h);
			p = lhc;
			goto reopen;
		} else {
			char *f = KERNEL$HANDLE_PATH(h);
			char *ff = strrchr(f, '/');
			int sl;
			if (__unlikely(!ff)) KERNEL$SUICIDE("is_exec: NO SLASH IN \"%s\"", f);
			sl = ff - f;
			p = alloca(sl + r + 2);
			memcpy(p, f, sl + 1);
			memcpy(p + sl + 1, lhc, r + 1);
			close(h);
			goto reopen;
		}
		c_enoexec:
		close(h);
		enoexec:
		errno = ENOEXEC;
		return 0;
	}
	close(h);
	if (__unlikely(LINK_CHECK_HEADER(&lh, (void *)&KERNEL$LIST_END, (void *)&KERNEL$LIST_END, (void *)&KERNEL$LIST_END, (void *)&KERNEL$LIST_END, (void *)&KERNEL$LIST_END, NULL))) {
		goto enoexec;
	}
	return 1;
#undef lhc
}

int is_cmd(const char *p)
{
	int l;
	if (__unlikely((l = strlen(p)) < 4)) {
		errno = ENOEXEC;
		return 0;
	}
	if (p[l - 4] != '.' || __upcasechr(p[l - 3]) != 'C' || __upcasechr(p[l - 2]) != 'M' || __upcasechr(p[l - 1]) != 'D') {
		errno = ENOEXEC;
		return 0;
	}
	return !open(p, O_RDONLY | _O_NOPOSIX | _O_CLOSE);
}

