#include <SPAD/DEV.H>
#include <SPAD/IOCTL.H>
#include <SPAD/SYNC.H>
#include <SPAD/SCSI.H>
#include <STDLIB.H>
#include <DLFCN.H>
#include <KERNEL/UNXCDROM.H>

static SCSI_PASS_THROUGH_HEAD *alloc_pass_through(size_t size)
{
	SCSI_PASS_THROUGH_HEAD *pt;
	size += sizeof(SCSI_PASS_THROUGH_HEAD);
	again:
	pt = memalign(BIO_SECTOR_SIZE, size);
	if (__unlikely(!pt)) {
		if (__likely(!KERNEL$MEMWAIT_SYNC(size))) goto again;
		return NULL;
	}
	memset(pt, 0, size);
	return pt;
}

static int do_command(int h, const __u8 *cmd, unsigned cmdlen, void *data, size_t datalen, int direction)
{
	int repeat_count = 0;
	IOCTLRQ io;
	SCSI_PASS_THROUGH_HEAD *pt = alloc_pass_through(datalen);
	if (__unlikely(!pt)) return -1;
	repeat:
	pt->direction = direction;
	pt->command_len = cmdlen;
	memcpy(pt->command, cmd, cmdlen);
	if (direction == PF_READ) memcpy(pt + 1, data, datalen);
	io.h = h;
	io.ioctl = IOCTL_BIO_SCSI_PASS_THROUGH;
	io.param = 0;
	io.v.ptr = (unsigned long)pt;
	io.v.len = sizeof(SCSI_PASS_THROUGH_HEAD) + datalen;
	io.v.vspace = &KERNEL$VIRTUAL;
	SYNC_IO_SIGINTERRUPT(&io, KERNEL$IOCTL);
	if (__unlikely(io.status < 0)) {
		errno = -io.status;
		free_ret:
		free(pt);
		return -1;
	}
	if (__unlikely(pt->scsi_status != SCSI_GOOD)) {
		int ra;
		void *scsi_dll;
		int (*recommended_action)(int status, __u8 scsi_status, __u8 *sense, unsigned sense_size);
		int (*sense_error_code)(__u8 scsi_version, __const__ SCSI_SENSE_DATA *sense, unsigned sense_size);
		int (*status_error_code)(__u8 scsi_status);
		if (__unlikely(!(scsi_dll = dlopen("SCSI.DLL", RTLD_NOW)))) {
			eio_free_ret:
			errno = EIO;
			goto free_ret;
		}
		if (!(recommended_action = dlsym(scsi_dll, "SCSI$RECOMMENDED_ACTION"))) {
			dlclose_eio_free_ret:
			dlclose(scsi_dll);
			goto eio_free_ret;
		}
		if (!(sense_error_code = dlsym(scsi_dll, "SCSI$SENSE_ERROR_CODE"))) {
			goto dlclose_eio_free_ret;
		}
		if (!(status_error_code = dlsym(scsi_dll, "SCSI$STATUS_ERROR_CODE"))) {
			goto dlclose_eio_free_ret;
		}
		ra = recommended_action(io.status, pt->scsi_status, pt->sense, SCSI_PASS_THROUGH_SENSE_LEN);
		if (ra == SCSI_ACTION_OK) {
			dlclose(scsi_dll);
			goto ok;
		}
		if (ra == SCSI_ACTION_RETRY_UNTIL_TIMEOUT || ra == SCSI_ACTION_RETRY_FEW_TIMES) {
			if (repeat_count < 1) {
				repeat_count++;
				goto repeat;
			}
		}
		if (pt->scsi_status == SCSI_CHECK_CONDITION) errno = -sense_error_code(2, (SCSI_SENSE_DATA *)pt->sense, SCSI_PASS_THROUGH_SENSE_LEN);
		else errno = -status_error_code(pt->scsi_status);
		dlclose(scsi_dll);
		goto free_ret;
	}
	ok:
	if (direction == PF_WRITE) memcpy(data, pt + 1, datalen);
	free(pt);
	return 0;
}

static void lba_to_msf(unsigned lba, __u8 *m, __u8 *s, __u8 *f)
{
	lba += CD_MSF_OFFSET;
	lba &= 0xffffff;
	*m = lba / (CD_SECS * CD_FRAMES);
	lba %= (CD_SECS * CD_FRAMES);
	*s = lba / CD_FRAMES;
	*f = lba % CD_FRAMES;
}

static unsigned msf_to_lba(unsigned m, unsigned s, unsigned f)
{
	return ((m * CD_SECS) + s) * CD_FRAMES + f - CD_MSF_OFFSET;
}

int cdrom_readtochdr(int h, struct cdrom_tochdr *tochdr)
{
	__u8 ret[0x0c];
	static const __u8 cmd[10] = { SCMD_CD_READ_TOC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00 };
	if (__unlikely(do_command(h, cmd, sizeof cmd, &ret, 0x0c, PF_WRITE))) return -1;
	tochdr->cdth_trk0 = ret[2];
	tochdr->cdth_trk1 = ret[3];
	return 0;
}

int cdrom_readtocentry(int h, struct cdrom_tocentry *tocentry)
{
	__u8 ret[0x0c];
	__u8 cmd[10] = { SCMD_CD_READ_TOC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00 };
	if (tocentry->cdte_format != CDROM_LBA && __unlikely(tocentry->cdte_format != CDROM_MSF)) {
		errno = EINVAL;
		return -1;
	}
	cmd[6] = tocentry->cdte_track;
	if (__unlikely(do_command(h, cmd, sizeof cmd, &ret, 0x0c, PF_WRITE))) return -1;
	tocentry->cdte_ctrl = ret[5] & 0x0f;
	tocentry->cdte_adr = ret[5] >> 4;
	tocentry->cdte_datamode = (tocentry->cdte_ctrl & 0x04) ? 1 : 0;
	tocentry->cdte_addr.lba = (ret[8] << 24) | (ret[9] << 16) | (ret[10] << 8) | ret[11];
	if (tocentry->cdte_format == CDROM_MSF) {
		lba_to_msf(tocentry->cdte_addr.lba, &tocentry->cdte_addr.msf.minute, &tocentry->cdte_addr.msf.second, &tocentry->cdte_addr.msf.frame);
	}
	return 0;
}

int cdrom_playmsf(int h, struct cdrom_msf *msf)
{
	__u8 cmd[10] = { SCMD_CD_PLAY_AUDIO_MSF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	cmd[3] = msf->cdmsf_min0;
	cmd[4] = msf->cdmsf_sec0;
	cmd[5] = msf->cdmsf_frame0;
	cmd[6] = msf->cdmsf_min1;
	cmd[7] = msf->cdmsf_sec1;
	cmd[8] = msf->cdmsf_frame1;
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_playtrkind(int h, struct cdrom_ti *ti)
{
	__u8 cmd[10] = { SCMD_CD_PLAY_AUDIO_MSF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	struct cdrom_tochdr tochdr;
	struct cdrom_tocentry tocentry0;
	struct cdrom_tocentry tocentry1;
	if (__unlikely(cdrom_readtochdr(h, &tochdr))) return -1;
	tocentry0.cdte_format = CDROM_LBA;
	tocentry0.cdte_track = ti->cdti_trk0;
	tocentry1.cdte_format = CDROM_LBA;
	tocentry1.cdte_track = ti->cdti_trk1 + 1;
	if (tocentry1.cdte_track > tochdr.cdth_trk1)
		tocentry1.cdte_track = CDROM_LEADOUT;
	if (__unlikely(cdrom_readtocentry(h, &tocentry0))) return -1;
	if (__unlikely(cdrom_readtocentry(h, &tocentry1))) return -1;
	if (__unlikely(tocentry0.cdte_addr.lba >= tocentry1.cdte_addr.lba)) {
		errno = EINVAL;
		return -1;
	}
	lba_to_msf(tocentry0.cdte_addr.lba, &cmd[3], &cmd[4], &cmd[5]);
	lba_to_msf(tocentry1.cdte_addr.lba - 1, &cmd[6], &cmd[7], &cmd[8]);
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_subchnl(int h, struct cdrom_subchnl *sc)
{
	__u8 ret[0x10];
	static const __u8 cmd[10] = { SCMD_CD_READ_SUB_CHANNEL, 0x02, 0x40, 0x01, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00 };
	if (sc->cdsc_format != CDROM_LBA && __unlikely(sc->cdsc_format != CDROM_MSF)) {
		errno = EINVAL;
		return -1;
	}
	if (__unlikely(do_command(h, cmd, sizeof cmd, &ret, 0x10, PF_WRITE))) return -1;
	sc->cdsc_audiostatus = ret[1];
	sc->cdsc_ctrl = ret[5] & 0xf;
	sc->cdsc_trk = ret[6];
	sc->cdsc_ind = ret[7];

	if (sc->cdsc_format == CDROM_MSF) {
		sc->cdsc_absaddr.msf.minute = ret[9];
		sc->cdsc_absaddr.msf.second = ret[10];
		sc->cdsc_absaddr.msf.frame = ret[11];
		sc->cdsc_reladdr.msf.minute = ret[13];
		sc->cdsc_reladdr.msf.second = ret[14];
		sc->cdsc_reladdr.msf.frame = ret[15];
	} else {
		sc->cdsc_absaddr.lba = msf_to_lba(ret[9], ret[10], ret[11]);
		sc->cdsc_reladdr.lba = msf_to_lba(ret[13], ret[14], ret[15]);
	}

	return -ENOSYS;
}

int cdrom_pause(int h)
{
	static const __u8 cmd[10] = { SCMD_CD_PAUSE_RESUME, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_resume(int h)
{
	static const __u8 cmd[10] = { SCMD_CD_PAUSE_RESUME, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_stop(int h)
{
	static const __u8 cmd[6] = { SCMD_START_STOP_UNIT, 0x01, 0x00, 0x00, 0x00, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_start(int h)
{
	static const __u8 cmd[6] = { SCMD_START_STOP_UNIT, 0x1, 0x0, 0x0, 0x1, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_eject(int h)
{
	static const __u8 cmd[6] = { SCMD_START_STOP_UNIT, 0x1, 0x0, 0x0, 0x2, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

int cdrom_close_tray(int h)
{
	static const __u8 cmd[6] = { SCMD_START_STOP_UNIT, 0x1, 0x0, 0x0, 0x3, 0x00 };
	return do_command(h, cmd, sizeof cmd, NULL, 0, 0);
}

